from odk_mailer.classes import Source, Fields, Message, Schedule
from odk_mailer.lib import globals, utils
import os
import json
from json import JSONEncoder
import hashlib
        
class MailjobEncoder(JSONEncoder):
    def default(self, o):
        return o.__dict__
    
class Mailjob:
    source: Source
    fields: Fields
    message: Message
    schedule: Schedule
    created: int

    hash: str
    json: str

    def __init__(hash_or_id = "") -> None:
        pass

    def create(
        self, 
        source: Source, 
        fields: Fields, 
        message: Message, 
        schedule: Schedule
        ) -> str:

        self.source = source
        self.fields = fields
        self.message = message
        self.schedule = schedule
        self.created = utils.now()

        self.setRecipients()
        self.setJSON()
        self.setHash()

        self.save()
        return self.hash

    def setRecipients(self):
        rows = self.source.get_rows()
        recipients = []
        relevant_fields = self.fields.data + [self.fields.email]
        
        for row in rows:   

            f_row = {k: row[k] for k in relevant_fields }
            recipients.append(f_row)

        self.recipients = recipients

    def save(self):
        # add to job json
        path_to_job = os.path.join(globals.odk_mailer_job, self.hash+'.json')
        with open(path_to_job, 'w', encoding='utf-8') as f:
            f.write(self.json)

        # add to jobs meta
        with open(globals.odk_mailer_jobs, "r+") as f:
            jobs_meta = json.load(f)

            new_job_meta = {
                "hash": self.hash,
                "scheduled": self.schedule.timestamp,
                "recipients": len(self.recipients),
                "state":0
            }
            jobs_meta.append(new_job_meta)
            f.seek(0)
            f.truncate()
            f.write(json.dumps(jobs_meta))
        
        return new_job_meta

    def dumpJSON(self):
        return json.dumps(vars(self), ensure_ascii=True, indent=4, cls=MailjobEncoder)

    def setJSON(self):
        self.json = self.dumpJSON()

    def setHash(self):
        self.hash = hashlib.sha256(self.json.encode()).hexdigest() 


from types import SimpleNamespace
import json
import os
from odk_mailer.lib import utils, globals

class Config():
    odk_host: str
    smtp_host: str
    smtp_port: str
    
    def __init__(self):
        if os.path.exists(globals.path_config):
            with open(globals.path_config, "r") as f:
                config = json.load(f, object_hook=lambda d: SimpleNamespace(**d))
            
            if not config:
                utils.abort("Config Error: Check .odk-mailer/config.json ")

            #required_keys = ["odk_host", "smtp_host", "smtp_port"]
            required_keys = ["smtp_host", "smtp_port"]

            for required_key in required_keys:
                if not required_key in vars(config):                
                    utils.abort(f"Invalid config: Key '{required_key}' is required.")

            self.odk_host =  config.odk_host

            self.smtp_host = config.smtp_host
            self.smtp_port = config.smtp_port

            # optional config parameters
            if hasattr(config, "smtp_user"):
                self.smtp_user = config.smtp_user
            if hasattr(config, "smtp_pass"):
                self.smtp_pass = config.smtp_pass


from odk_mailer.lib import utils

class Message:
    sender: str
    subject: str
    source: str
    content: str
    format: str

    def __init__(self, data: []) -> None:

        self.sender = data["sender"]
        self.subject = data["subject"]
        self.source = data["source"]
        self.format = data["type"]
        
        if self.source == "stdin":
            self.content = data["location"]

        if self.source == "path":            
            with open(data["location"]) as f:
                data_string = f.read()
                base64_string = utils.base64_encode_str(data_string)
                self.content = base64_string

        if self.source == "url":
            #tbd: read from url and store as base64
            self.content = data["location"]
                    


from pathlib import Path
from odk_mailer.lib import globals
import typer
import json

def init():

  # create base dir if not exists
  path_base = Path(globals.odk_mailer_base)
  path_base.mkdir(parents=True, exist_ok=True)
  
  # create config.json if not exists
  path_config  =  Path(globals.path_config)
  if not path_config.exists():
      path_config.touch()

  if path_config.stat().st_size == 0:
      path_config.write_text("{}")

  if not is_json(path_config.read_text()):
      raise typer.Exit("The config.json file is invalid.\n" + path_config)        

  # create job dir if not exists
  path_job = Path(globals.odk_mailer_job)
  path_job.mkdir(parents=True, exist_ok=True)

  # create jobs.json if not exists
  path_jobs = Path(globals.odk_mailer_jobs)

  if not path_jobs.exists():
      path_jobs.touch()

  if path_jobs.stat().st_size == 0:
      path_jobs.write_text("[]")

  if not is_json(path_jobs.read_text()):
      raise typer.Exit("The jobs.json file is invalid.\n" + path_jobs)
  

def is_json(myjson):
  try:
    json.loads(myjson)
  except ValueError as e:
    return False
  return True


from email.message import EmailMessage
import smtplib
import time

def send_test_mail(sender, recipient, host, port):

    subject = "Test Mail"
    message = "This is a test mail. Time: " + str(int(time.time()))

    email = EmailMessage()
    email['Subject'] = subject
    email['From'] = sender
    email['To'] = recipient
    email.set_content(message, subtype="plain")

    try:
        smtp = smtplib.SMTP(timeout=5)
        smtp.set_debuglevel(2)
        smtp.connect(host, port)
        smtp.send_message(email)
        smtp.quit()
        print("Successfully sent email to " + email["To"])
    except Exception as error:
        print(error)



import os
from types import SimpleNamespace
import smtplib
from email.message import EmailMessage
from odk_mailer.lib import globals, utils
import json
import sys

# return unformatted string instead of raising error
# when key is missing within dictionary
# https://stackoverflow.com/a/17215533/3127170
class SafeDict(dict):
    def __missing__(self, key):
        return '{' + key + '}' 
    

class Mailer:
    hash: str
    sender:str
    format: str
    content: str
    recipients: []

    subject: str
    body: str
    sender: str
    headers: str

    def __init__(self, hash: str, dry: bool, verbose:bool, config):
        self.hash = hash
        self.verbose = verbose
        self.config = config
        self.dry = dry

        with open(os.path.join(globals.odk_mailer_job, self.hash+'.json'), 'r', encoding='utf-8') as f:
            job = json.load(f, object_hook=lambda d: SimpleNamespace(**d))

        self.message = job.message
        self.recipients = job.recipients

    def send(self):

        if self.verbose:
            print()
            print(f"Run Mailjob {self.hash}")
            print("============================================================================")
            if self.dry:
                print("Dry Run enabled. No Emails will be sent.")
            else:
                print("Sending emails..")
            print("============================================================================")



        # Decode base64 if needed
        if self.message.source == "stdin":
            text = self.message.content
        if self.message.source in ["path", "url"]:
            text = utils.base64_decode_str(self.message.content)

        if self.dry and self.verbose:
            print()
            print("Message Summary")
            print("-----------------------------------------")
            print(f"sender: {self.message.sender}")
            print(f"subject: {self.message.subject}")
            print(f"source: {self.message.source}")
            print(f"format: {self.message.format}")
            print(f"# recipients: {len(self.recipients)}")
            print("-----------------------------------------")
            print()

        idx=0
        for recipient in self.recipients:            
            if self.dry:
                print(f"#{idx}")
                print(f"recipient.email{recipient.email}")
                print(text.format_map(SafeDict(vars(recipient))))
            else: 
                print()
                print(f"(#{idx+1}) Attempting to send.. ")
                success = self.smtp(
                    recipient.email, 
                    text.format_map(SafeDict(vars(recipient)))
                )
            idx = idx + 1

        # update job state: pending, success, errors


    def smtp(self, recipient, message, type='plain'):

        email = EmailMessage()
        email['Subject'] = self.message.subject
        email['From'] = self.message.sender
        email['To'] = recipient
        email.set_content(message, subtype=type)

        try:            
            smtp = smtplib.SMTP(timeout=5)
            if self.verbose:
            # enable debugging by CLI flag --debug
                smtp.set_debuglevel(2)
            smtp.connect(self.config.smtp_host, self.config.smtp_port)

            if hasattr(self.config, 'smtp_user') and hasattr(self.config, 'smtp_pass'):
                smtp.login(self.config.smtp_user, self.config.smtp_pass)
            # if username and password are supplied, perform smtp.login()
            # requires additional actions, such as setting TLS or SSL
            smtp.send_message(email)
            smtp.quit()
            # write into /log/timestamp_<hash>.log
            # log.write("Successfully sent email to " + email["To"])
            return True
        except Exception as error:
            # write into /log/timestamp_<hash>.log
            # raise exception to interrupt loop
            print(error)
            print("Failed sending mail to: " + email['To'])
            print()
            return False


import os

odk_mailer_base = os.path.join(os.getenv("HOME"), ".odk-mailer")
odk_mailer_jobs = os.path.join(odk_mailer_base, "jobs.json")
odk_mailer_job = os.path.join(odk_mailer_base, "job")
path_config = os.path.join(odk_mailer_base, "config.json")


import time
from datetime import datetime

class Schedule:
    now: bool   # can be removed in future
    timestamp: int

    def __init__(self, data: []):
        self.now = data["now"]        

        if self.now:
            self.timestamp = int(time.time())

        else:
            _datetime = datetime.fromisoformat(data["datetime"])
            self.timestamp = int(datetime.timestamp(_datetime))


class Fields:
    email: str
    data: []

    def __init__(self, data: []):
            
        self.email = data["email"]
        if "data" in data:
            self.data = data["data"]
        else:
            self.data = []


import os
import json
from odk_mailer.lib import parser, prompts, utils, globals, smtp
from odk_mailer.classes.mailer import Mailer
from odk_mailer.classes.config import Config
from odk_mailer.classes.Source import Source
from odk_mailer.classes.Fields import Fields
from odk_mailer.classes.Message import Message
from odk_mailer.classes.Schedule import Schedule
from odk_mailer.classes.Mailjob import Mailjob

def create(input_source, input_fields, input_message, input_schedule):
   
    # source
    if not input_source:
        source_data = prompts.prompt_source()
    else: 
        source_data = parser.parse_source(input_source)
    source = Source(source_data)

    # fields
    headers = source.get_headers()
    if not input_fields:
        fields_data = prompts.prompt_fields(headers)
    else: 
        fields_data = parser.parse_fields(input_fields, headers)
    fields = Fields(fields_data)

    # message
    if not input_message:
        message_data = prompts.prompt_message()
    else:
        message_data = parser.parse_message(input_message)
    message = Message(message_data)

    # schedule
    if not input_schedule:
        schedule_data = prompts.prompt_schedule()
    else:
        schedule_data = parser.parse_schedule(input_schedule)
    schedule = Schedule(schedule_data)        

    # Create Mailjob
    mailjob = Mailjob()
    created = mailjob.create(source, fields, message, schedule)
    print(created)
    
    ###
    # Features
    ###

    # 1. Reminders 
    # tbd: Reminders
    # reminders have two attributes:
    # total_amount, frequency, e.g. 3 times in total, every hour|day|week|custom frequency
    # after first scheduled send

    # reminders will be stored inside .odk-mailer/reminder/hash_[instance_id].json
    # having updated recipients from non-respondents (calculated from base recipients and respondents) 
    # on a per reminder case

    # https://github.com/tertek/zapier-odk-scripts/blob/main/get_non_responding.py
    # reminders require api connection and following inputs
    # form_register = "test_form_register" #name of form that is used for registration, given as api or csv
    # form_follow = "test_form_follow"  # name of form that is used for follow up
    # # if use_form_attachment
    # form_follow_attached = "test_form_follow_attached" # name of form that is used for follow up; 
    # form_attachment = "follow.csv" # name of form attachment attached to form_follow; 
    # # field config
    # field_email_register = "email_register" # name of email field for registration form, given
    # field_email_follow = "email_follow"
    # field_email_follow_attach = "email"

    # 2. Cron Jobs
    # add cron-job via https://pypi.org/project/python-crontab/ and https://stackabuse.com/scheduling-jobs-with-python-crontab/
    # We will can use a single job, that checks every hour if we have pending Mailjobs, running 'odk-mailer evaluate
    # process reminder tasks as follows: 
    # 1. Perform API request to calculate reminder recipients
    # 2. Send Emails based on calculated recipients
    # 3. Summarize progress

    # 3. Recipients Validation
    # Validate recipients and ignore (e.g. filter) invalid emails to proceed
    # code example
    # if not recipients.validate(email_field):
    #     utils.render_table(["id", "email_field", "error"], recipients.invalidEmails)
    #     ignore_invalid_emails = typer.confirm("Invalid emails found. Would you like to continue although you have invalid emails?")
    
    #     if not ignore_invalid_emails:
    #         raise typer.Exit("\nAborted.")

    # with typer.progressbar(self.data) as progress:
    #     for row in progress:
    #         total += 1
    #         email = row[email_field]
    #         try:
    #             if not email:
    #                 raise EmailNotValidError("Email address missing. Check for missing delimiters (',') in your CSV file.")
    #             # Disable DNS checks since this can be blocked for unknown reasons within network.
    #             validate_email(email, check_deliverability=True)
    #         except EmailNotValidError as e:
    #             invalid = [str(total) ,email, str(e)]
    #             invalid_emails.append(invalid)

def run(hash_or_id, dry=False, verbose=False):

    odk_mailer_config = Config()

    if not odk_mailer_config:
        utils.abort("Cannot send emails without Config")

    if not hash_or_id:
        utils.abort("ID/Hash is required")

    with open(globals.odk_mailer_jobs, "r") as f:
        jobs = json.load(f)

    found = next((obj for obj in jobs if obj["hash"].startswith(hash_or_id)), None)
    if not found:
        utils.abort("Job not found.")

    hash = found['hash']
     
    # check if ready to be sent
    # simple check: scheduled <= now
    # advan. check: hasReminders AND UnsentReminderTime <= now
    
    # if found["scheduled"] > utils.now():
    #     utils.abort("Schedule is in future")

    mailer = Mailer(hash, dry, verbose, odk_mailer_config)
    mailer.send()
    # in case we have a reminder case, generate reminder contents from reminders/hash_reminderId.json

def delete(hash):
    if not hash:
            utils.abort("ID is required")

    with open(globals.odk_mailer_jobs, "r") as f:
        jobs = json.load(f)

    found = next((obj for obj in jobs if obj["hash"].startswith(hash)), None)

    if not found:
        utils.abort("Job not found.")
   
    # tbd: confirm
    
    # deletion from /job/<hash>.json
    path_job = os.path.join(globals.odk_mailer_job, found['hash']+'.json')
    if os.path.exists(path_job):
        os.remove(path_job)

    # deletion from /jobs.json
    jobs_updated = list(filter(lambda x: x['hash']!=found['hash'], jobs))   
    
    with open(globals.odk_mailer_jobs, "w") as f:
        f.write(json.dumps(jobs_updated))

    print("Deleted " + found['hash'])    

def list_jobs():

    with open(globals.odk_mailer_jobs, "r+") as f:
        jobs = json.load(f)

    if len(jobs) > 0:
        utils.print_jobs(jobs)

    else: print("There are no mailjobs yet. Create one to see it listed.")


def evaluate(dry=False):

    with open(globals.odk_mailer_jobs, "r+") as f:
        jobs = json.load(f)

    evals = []

    for job in jobs:
        if job["scheduled"] <= utils.now():
            # simple evaluation: check if scheduled time is smaller/equal to now
            # if true, add to selected list
            evals.append(job["hash"])

        ###
        # untested code, since reminders are not yet implemented in create command
        elif "reminders" in job and len(job["reminders"]) > 0:
            # advanced evaluation: addtionally check if job has reminder times that are smaller/equal to now
            # if true, add to selected list
            print("Addiitonally checking if we have any valid reminders")
            for reminder in job["reminders"]:
                if reminder["timestamp"] <= utils.now():
                    evals.append(job["hash"])
                    break
        ###
                
        else:
            # skipping this job since not qualified to be run
            pass

    if dry:
        print(evals)
        print(len(evals))

    else:
        for eval in evals:
            run(eval)

def test(sender, recipient, host, port):
    print()
    print(f"Sending test mail from {sender} to:  {recipient} via: {host}:{port}")
    print()
    smtp.send_test_mail(sender, recipient, host, port)


def db():
    print("Testing db")


from odk_mailer.lib import utils, validators
import os

def parse_source(input):

    source_list = input.split("::", 6)
    source_count = len(source_list)
    
    # --source=file::tests/data/valid.csv
    # --source=url::https://some.url.tld/file.csv
    if source_count == 2:

        type, location = source_list

        if type not in ["path", "url"]:
            utils.abort("Invalid source type, expected 'path' or 'url.")

        if type == "path":
            ext = os.path.splitext(location)[-1].lower()
            # invalid file extension
            if not ext == ".csv":
                utils.abort("Invalid file extension, expected '.csv'")
            # invalid file path
            if not os.path.isfile(location):
                utils.abort("Invalid file path, expected existing file")

            return {
                'type': type,
                'location': location
            }
                    
        if type == "url":
            utils.abort("URL not yet supported!")

            return {
                'type': type,
                'location': location
            }            

    if source_count == 6:
        utils.abort("API not yet supported!")

        # type, form, project, host, user, pass = source_list
        # check API connection & credentials
        # promptAPIcredentials() // host, user, pass, project
        # checkAPIconnection()  // auth endpoint

        # return {
        #     'type': type,
        #     'location': form,
        #     'api_proj': project,
        #     'api_host': host,
        #     'api_user': user,
        #     'api_pass': pass
        # }        

    else:
        utils.abort("Invalid file/api source string")


def parse_fields(input, headers):

    fields_list = input.split("::", 2)
    fields_count = len(fields_list)

    if not fields_count in [1,2]:        
        utils.abort("Invalid email/data fields string")
    
    email = fields_list[0]

    if email not in headers:
        utils.abort(f"Invalid email field '{email}'.")

    if fields_count == 1:
        return {
            'email': email
        }
    
    if fields_count == 2:
        data = fields_list[1].split(",") 
        for field in data:
            if not field in filter(lambda x: x != email, headers):
                utils.abort(f"Invalid data field '{field}'")
    
        return {
            'email': email,
            'data': data
        }
    
def parse_message(input):
    message_list = input.split("::", 5)
    message_count = len(message_list)

    if not message_count == 5:
        utils.abort("Define message as \[sender:str]::\[subject:str]::\[source:'stdin'|'path'|'url']::\[location:str]::\[type:'html'|'text']")

    sender, subject, source, location, type = message_list

    if not validators.email_address("", sender):        
        raise utils.abort("Invalid message sender. Use valid email address.")   
    
    if source not in ["stdin", "path", "url"]:
        raise utils.abort("Invalid message source. Use either 'stdin' or 'path or 'url.")
    
    if type not in ["text", "html"]:
        raise utils.abort("Invalid message type. Use either 'plain' or 'html'.")    

    if source == "path":
        if not os.path.isfile(location):
            utils.abort("Invalid file path, expected existing file")        
    
    return {
        'sender': sender,
        'subject': subject,
        'source': source,
        'location': location,
        'type': type
    }

def parse_schedule(input):    
    if input == "now":
        now = True
        datetime = None
    else: 
        now = False
        datetime = input

    return {
        'now': now,
        'datetime': datetime
    }


import typer
from typing import Optional
from typing_extensions import Annotated
from odk_mailer import commands, before

app = typer.Typer(
    add_completion=False, 
    pretty_exceptions_enable=True
)

@app.callback()
def callback():
    """
    ODK Mailer

    Setup mail jobs by fetching recipients from CSV files or ODK API.

    Run mail jobs immediately or schedule them to be run over time.

    Evaluate available Mailjobs.

    """

    before.init()

@app.command()
def create(
    source: Annotated[Optional[str], typer.Option("--source", "-s", help="Define source as [type:path|url|api]::[location:str]")]= "",
    fields: Annotated[Optional[str], typer.Option("--fields", "-f", help="Define fields as [email:str]::[data: 'field_1,field_2']")]= "",
    message: Annotated[Optional[str], typer.Option("--message", "-m", help="Define message as [sender:str]::[subject:str]::[source:'stdin'|'path'|'url']::[location:str]::[type:'html'|'text']")]= "",
    schedule: Annotated[Optional[str], typer.Option("--schedule", help="Define schedule as 'now' or [time:str] in 'YYYY-MM-DD HH:mm' format")]= "",
):
    """
    Create mail job
    """
    commands.create(source, fields, message, schedule)

@app.command()
def run(
    id: Annotated[str, typer.Argument(help="Hexadecimal hash")],
    dry: Annotated[bool, typer.Option("--dry", help="Dry run without sending mails.")] = False,
    verbose: Annotated[bool, typer.Option("--verbose", help="Print out smtp debugging information")] = False,
):
    """
    Run mail job
    """
    commands.run(id, dry, verbose)

@app.command()
def delete(
    id: Annotated[str, typer.Argument(help="Hexadecimal hash")]
):
    """
    Delete mail job
    """
    commands.delete(id)

@app.command()
def list():
    """
    List mail jobs
    """    
    commands.list_jobs()

@app.command()
def evaluate(
    dry: Annotated[bool, typer.Option("--dry", help="Dry run without sending mails.")] = False
):
    """
    Evaluate mail jobs
    """
    commands.evaluate(dry)

@app.command()
def test(
    sender: Annotated[str, typer.Option("--sender")] = "odk-mailer@freesmtpservers.com",
    recipient: Annotated[str, typer.Option("--recipient")] = "foo@bar.com",
    host: Annotated[str, typer.Option("--host")] = "smtp.freesmtpservers.com",
    port: Annotated[str, typer.Option("--port")] = "25"
):
    """
    Send test mail
    """
    commands.test(sender, recipient, host, port)



import csv
class Source:
    type: str
    location: str
    project: int    
    hostname: str
    username: str
    password: str

    def __init__(self, data: []):

        self.type = data["type"]

        if self.type == "path":
            self.location = data["location"]
        
        if self.type == "url":
            self.location = data["location"]

        if self.type == "api":
            self.location = data["location"]
            self.project = data["api_proj"]
            self.hostname = data["api_host"]
            self.username = data["api_user"]
            self.password = data["api_pass"]

    def get_headers(self):            
        if self.type == 'path':        
            with open(self.location, newline='') as f:
                reader = csv.DictReader(f, skipinitialspace=True)
                headers = reader.fieldnames
            return headers
        
        

    def get_rows(self):
        if self.type == 'path':        
            with open(self.location, newline='') as f:
                reader = csv.DictReader(f, skipinitialspace=True)
                #headers = reader.fieldnames
                rows = []
                for row in reader:
                    rows.append(row)

            return rows
        
        raise Exception("Source type is not yet implemented" )
       


# import logging
# logging.basicConfig(
#     filename='example.log', 
#     encoding='utf-8', 
#     level=logging.DEBUG, 
#     format='%(asctime)s: %(levelname)s: %(message)s'
# )

# def write(msg, type="info"):

#     if type == "info":
#         logging.info(msg)